
/* Photographic light meter using TSL2591 Digital Light Sensor and HD44780 LCD display
  Dynamic Range: 600M:1
  Maximum Lux: 88K
  Conversion to whole integer seconds or reciprocals for shutter speeds <1 is tricky
  Here I used a float array of possible shutter speeds for the calculations (shutter_float[])
  but also a synchronised integer array of shutter speeds populated with
  natural numbers down to 1 second then and reciprocals for values less than 1 second (shutter[])
*/

#include <Wire.h>
#include <Adafruit_Sensor.h>
#include "Adafruit_TSL2591.h"
#include <hd44780.h>                       // main hd44780 header
#include <hd44780ioClass/hd44780_I2Cexp.h> // i2c expander i/o class header
#include <Button.h>  //uses the button library for easy button control

// connect SCL of light sensor and LCD display to I2C Clock (A5 on Nano and Uno)
// connect SDA of light sensor and LCD display to I2C Data (A4 on Nano and Uno)
// connect momentary switches to pins D3, D4, D5, D6 for ISO up/down and Aperture up/down resp
// connect Vin to 3.3-5V DC
// connect GROUND to common ground

//int iso_stops = 11;
//int aperture_stops = 8;
//int shutter_stops = 24;
const int LCD_COLS = 16;
const int LCD_ROWS = 2;
const float aperture[8] = {2.0, 2.8, 4.0, 5.6, 8.0, 11.0, 16.0, 22.0};
const int shutter[24] = {1000, 500, 250, 120, 60 , 30, 16, 8, 4, 2, 1, 2, 4, 8, 15, 30, 60, 125, 250, 500, 1000, 2000, 4000, 8000};
const float shutter_float[24] = {1000, 500, 250, 120, 60, 30, 16, 8, 4, 2, 1, 0.5, 0.25, 0.125, 0.067, 0.033, 0.017, 0.008, 0.004, 0.002, 0.001, 0.0005, 0.00025, 0.000125};
//int ISO[iso_stops] = {12, 25, 50, 100, 125, 160, 200, 250, 320, 400, 500, 640, 800, 1000, 1250, 1600, 2000, 2500, 3200, 4000, 5000, 6400, 8000, 10000, 12800};
const float ISO[11] = {12, 25, 50, 100, 200, 400, 800, 1600, 3200, 6400, 12800};
int aperture_index = 5; // at switch on aperture is set to f8
int shutter_index = 0; // at switch on shutter is set to 1000 sec, this will be altered automatically after the first reading
int ISO_index = 4; // at switch on ISO is set to 100
float lux;
//int integration = 100; changes in integration time not implemented at this time
float t; //calculated shutter speed before rounding, later used for rounded shutter speed
float ISO_float = 100;
float aperture_float = 5.6;
float calibration_factor = 1.56;
char sens = 'L';

Button iso_down(4); // Connect button between pin 4 and GND
Button iso_up(3); // Connect button between pin 3 and GND
Button aperture_down(6); // Connect button between pin 6 and GND
Button aperture_up(5); // Connect button between pin 5 and ground
Button short_long_integration(6);
//Button button6(7); never know when you might need another one

Adafruit_TSL2591 tsl = Adafruit_TSL2591(2591); // pass in a number for the sensor identifier (for your use later)


hd44780_I2Cexp lcd; // automatic detection i2c address for the lcd
int status;


// set the lcd rows and columns and check the procedure worked OK
void lcd_setup()
{
  status = lcd.begin(LCD_COLS, LCD_ROWS);
  if (status) // non zero status means it was unsuccesful
  {
    status = -status; // convert negative status value to positive number

    // begin() failed so blink error code using the onboard LED if possible
    hd44780::fatalError(status); // does not return
  }
}

// initalization was successful, the backlight should be on now



/**************************************************************************/
/*
    Displays some basic information on this sensor from the unified
    sensor API sensor_t type (see Adafruit_Sensor for more information)
*/
/**************************************************************************/
void displaySensorDetails(void)
{
  sensor_t sensor;
  tsl.getSensor(&sensor);
  Serial.println(F("------------------------------------"));
  Serial.print  (F("Sensor:       ")); Serial.println(sensor.name);
  Serial.print  (F("Driver Ver:   ")); Serial.println(sensor.version);
  Serial.print  (F("Unique ID:    ")); Serial.println(sensor.sensor_id);
  Serial.print  (F("Max Value:    ")); Serial.print(sensor.max_value); Serial.println(F(" lux"));
  Serial.print  (F("Min Value:    ")); Serial.print(sensor.min_value); Serial.println(F(" lux"));
  Serial.print  (F("Resolution:   ")); Serial.print(sensor.resolution, 4); Serial.println(F(" lux"));
  Serial.println(F("------------------------------------"));
  Serial.println(F(""));
}



/**************************************************************************/
/*
    Configures the gain and integration time for the TSL2591
*/
/**************************************************************************/
void configureSensor(void)
{
  // You can change the gain on the fly, to adapt to brighter/dimmer light situations
  tsl.setGain(TSL2591_GAIN_LOW);    // 1x gain (bright light)
  //tsl.setGain(TSL2591_GAIN_MED);      // 25x gain
  //tsl.setGain(TSL2591_GAIN_HIGH);   // 428x gain
  //tsl.setGain(TSL2591_GAIN_MAX);   // 9876x gain

  // Changing the integration time gives you a longer time over which to sense light
  // longer timelines are slower, but are good in very low light situtations!
  //tsl.setTiming(TSL2591_INTEGRATIONTIME_100MS);  // shortest integration time (bright light)
  // tsl.setTiming(TSL2591_INTEGRATIONTIME_200MS);
  tsl.setTiming(TSL2591_INTEGRATIONTIME_300MS);
  // tsl.setTiming(TSL2591_INTEGRATIONTIME_400MS);
  // tsl.setTiming(TSL2591_INTEGRATIONTIME_500MS);
  // tsl.setTiming(TSL2591_INTEGRATIONTIME_600MS);  // longest integration time (dim light)

  /* Display the gain and integration time for reference sake */
  Serial.println(F("------------------------------------"));
  Serial.print  (F("Gain:         "));
  tsl2591Gain_t gain = tsl.getGain();
  switch (gain)
  {
    case TSL2591_GAIN_LOW:
      Serial.println(F("1x (Low)"));
      break;
    case TSL2591_GAIN_MED:
      Serial.println(F("25x (Medium)"));
      break;
    case TSL2591_GAIN_HIGH:
      Serial.println(F("428x (High)"));
      break;
    case TSL2591_GAIN_MAX:
      Serial.println(F("9876x (Max)"));
      break;
  }
  Serial.print  (F("Timing:       "));
  Serial.print((tsl.getTiming() + 1) * 100, DEC);
  Serial.println(F(" ms"));
  Serial.println(F("------------------------------------"));
  Serial.println(F(""));
}


/**************************************************************************/
/*
    Program entry point for the Arduino sketch
*/
/**************************************************************************/
void setup(void)
{
  // some initiation data printing to teh serial monitor
  Serial.begin(9600);
  Serial.println(F("Starting Adafruit TSL2591 Test!"));
  if (tsl.begin())
  {
    Serial.println(F("Found a TSL2591 sensor"));
  }
  else
  {
    Serial.println(F("No sensor found ... check your wiring?"));
    while (1);
  }
  // put the non-volatile titles on the LCD screen
  status = lcd.begin(LCD_COLS, LCD_ROWS);
  lcd.clear();
  lcd.leftToRight();
  lcd.setCursor(0, 0);
  //lcd.print("Lux=");
  lcd.setCursor(7, 0);
  lcd.print("iso=");
  lcd.setCursor(0, 1);
  lcd.print("f=");
  lcd.setCursor (6, 1);
  lcd.print("t=");

  /* Display some basic information on this sensor */
  displaySensorDetails();

  /* Configure the sensor */
  configureSensor();

  //configure buttons
  iso_down.begin();
  iso_up.begin();
  aperture_down.begin();
  aperture_up.begin();
  short_long_integration.begin();




  // Now we're ready to get readings ... move on to loop()!
}


/**************************************************************************/
/*
         Read IR and Full Spectrum at once and convert to lux
*/
/**************************************************************************/
void advancedRead(void)
{
  // Read 32 bits with top 16 bits IR, bottom 16 bits full spectrum
  // At every read the gain is first set to LOW, the gain may then be increased
  // depeinding on the value of lux. "ir" and "full" are also tested because
  // if these go to zero an error message is reported for lux.

  tsl.setGain(TSL2591_GAIN_LOW);    // 1x gain (bright light)
  uint32_t lum = tsl.getFullLuminosity(); // Declare 32 bit variable and get luminosity
  uint16_t ir, full; // declare 16 bit variables for infra-red and full spectrum raw data
  ir = lum >> 16; full = lum & 0xFFFF; // Split lum up, it contains the IR and full raw data
  lux = tsl.calculateLux(full, ir); // ask the unit to calculate Lux fom the raw data
  sens = 'L';
  Serial.print(" Gain: "); tsl2591Gain_t gain = tsl.getGain(); Serial.println(gain);

  // Now test the LUX value and set the gain appropriately via a sequence of if's

  if (lux < 100 || ir < 1 || full < 1)
  {
    tsl.setGain(TSL2591_GAIN_MED);      // 25x gain
    lum = tsl.getFullLuminosity();
    ir = lum >> 16;
    full = lum & 0xFFFF;
    lux = tsl.calculateLux(full, ir);
    sens = 'M';
    Serial.print(" Gain: "); gain = tsl.getGain(); Serial.println(gain);
  }

  if (lux < 10 || ir < 1 || full < 1)
  {
    tsl.setGain(TSL2591_GAIN_HIGH);      // 428x gain
    lum = tsl.getFullLuminosity();
    ir = lum >> 16; full = lum & 0xFFFF;
    lux = tsl.calculateLux(full, ir);
    Serial.print(" Gain: "); gain = tsl.getGain(); Serial.println(gain);
    sens = 'H';
  }

  if (lux < 1 || ir < 1 || full < 1)
  {
    tsl.setGain(TSL2591_GAIN_MAX);      // 9876x gain
    lum = tsl.getFullLuminosity();
    ir = lum >> 16; full = lum & 0xFFFF;
    lux = tsl.calculateLux(full, ir);
    Serial.print(" Gain: "); gain = tsl.getGain(); Serial.println(gain);
    sens = 'X';
  }



  lux = lux * calibration_factor; // Adjust the reading, required if a filter or lens is placed over the sensor

  // Print the readings to the serial monitor

  Serial.print(F("[ ")); Serial.print(millis()); Serial.print(F(" ms ] "));
  Serial.print(F("IR: ")); Serial.print(ir);  Serial.print(F("  "));
  Serial.print(F("Full: ")); Serial.print(full); Serial.print(F("  "));
  Serial.print(F("Visible: ")); Serial.print(full - ir); Serial.print(F("  "));
  Serial.print(F("Lux: ")); Serial.println(lux), 6;
  //  Serial.print(" Gain: "); tsl2591Gain_t gain = tsl.getGain(); Serial.println(gain);

}

/**************************************************************************/
void checkbuttons()
{

  //check if the operator has pressed any buttons to change ISO or Aperture

  calc(); // so the operator has something to look at before pressing buttons

  if (iso_down.pressed())  //change ISO and display results
  { ISO_index --;
    if (ISO_index == -1) ISO_index = 0;
    ISO_float = ISO[ISO_index];
    calc();
  }

  if (iso_up.pressed()) //change ISO and display results
  { ISO_index ++;
    if (ISO_index == 11) ISO_index = 10;
    ISO_float = ISO[ISO_index];
    calc();
  }

  if (aperture_down.pressed())//change aperture and display results
  { aperture_index --;
    if (aperture_index == -1) aperture_index = 0;
    aperture_float = aperture[aperture_index];
    calc();
  }

  if (aperture_up.pressed())  //change aperture and display results
  { aperture_index ++;
    if (aperture_index == 8) aperture_index = 7;
    aperture_float = aperture[aperture_index];
    calc();
  }
}

/***************************************************************************/
void calc() //uses lux, iso, and aperture to calculate shutter speed and EV
{
  //first calculate EV for iso 100
  float EV;
  EV = (log10(lux) - 0.39794) / 0.30103; //this is equiv to (log(lux)-log(1/2.5))/log(2)
  //EV=(log(lux)- log(2/5))/log(2);

  //Now calculate the shutter speed for given EV and aperture
  t = (1 / pow(2, (EV - 1))) * aperture_float * aperture_float * 50 / ISO_float;

  // Now the hard bit, we have to quantise the shutter speed to one of the
  // values found on a camera and also convert results to reciprocals for
  // shutter speeds less than 1 sec.

  shutter_index = 0;  //this is a pointer to both arrays shutter_float[] and shutter[]
  //shutter_float[] is used in the calculation and contains an regular series decimal shutter speeds
  //shutter[] is an integer array and is used only for displaying the shutter speeds in the fashion that they appear on a camera
  //in the calculation shutter_index is derived from Shutter_float[], then that is mapped across to shutter[] for display
  
  while (t < shutter_float[shutter_index])
  {
    shutter_index++;
  }
  //print some results to the monitor
  //  Serial.print("shutter "); Serial.print(shutter_float[shutter_index - 1], 5); Serial.print(" - "); Serial.print(shutter_float[shutter_index], 5);
  //  Serial.print(" t= "); Serial.println(t, 5);

  // Find what proportion is left over (interpolate) so we can indicate a part stop
  float rem = -(t - shutter_float[shutter_index - 1]) / (shutter_float[shutter_index - 1] - shutter_float[shutter_index]);
  //  Serial.print("Remainder = "); Serial.print(rem);

  //Change t to the nearest of adjacent stops
  if (rem < .5) shutter_index = shutter_index - 1;
  t = shutter_float[shutter_index];


  // Finally print results to LCD

  //put volatile content on the LCD
  lcd.setCursor(0, 0);
  lcd.print("       ");
  lcd.setCursor(0, 0);
  lcd.print(lux);
  lcd.setCursor(11, 0);
  lcd.print("     ");
  lcd.setCursor(11, 0);
  lcd.print(int(ISO_float));
  lcd.setCursor(2, 1);
  lcd.print(aperture_float);
  /*at high values the iso is partially overwritten by the Lux figure which can be 8 characters long
    so blank out 2 spaces and write the i from iso back in*/
  lcd.setCursor(5, 0); lcd.print("  i");  //
  lcd.setCursor (5, 1);
  lcd.print(" t=       ");
  lcd.setCursor(8, 1);
  if (t >= 1) lcd.print(shutter[shutter_index]);
  else {
    lcd.print("1/");
    lcd.print(shutter[shutter_index]);
    lcd.print("    ");
  }
  lcd.setCursor(15, 2);
  lcd.print(sens);
}

/**************************************************************************/
/*
    Arduino loop function, called once 'setup' is complete (your own code
    should go here)
*/
/**************************************************************************/
void loop(void)
{
  advancedRead();
  checkbuttons();
  //delay(1000);
}
